define(
  'structure/testy/ManageStatuses',
  [
    'jquery',
    'underscore',
    'almworks/util/Util',
    'jira/util/formatter',
    'aui/message'
  ],
  function ($, _, U, Formatter, Message) {
    'use strict';

    var TestyUtils = window.namespace('almworks.testy.utils');
    var manageStatusesTemplates = window.namespace('almworks.testy.templates.statuses');

    function ManageStatuses() {
      // Update RestTestStatus doc if color set changes
      var AVAILABLE_SETS = ['green', 'red', 'yellow', 'grey', 'invis', 'blue', 'sky-blue', 'brown', 'purple', 'pink'];
      var NEW_ID = 'newStatus';
      var MESSAGE_CONTAINER_ID = '#testy-message-bar';
      var SORTABLE_OPTIONS = {
        axis: 'y',
        handle: '.st-dh',
        placeholder: 'tty-status-placeholder',
        start: function (event, ui) {
          ui.item.addClass('tty-status-dragging');
        },
        stop: function (event, ui) {
          ui.item.removeClass('tty-status-dragging');
          reorder(ui.item);
        }
      };

      var statusMap = {};
      var userStatusNameMap = {};
      var colorWidth = new U.Cell(0).pipeChanges(propagateColorWidth);
      var updateTextWidthLifecycle = new U.Lifecycle();
      var newStatusColorCell = null;
      var newStatusCell = new U.Cell(null);
      var creatingNew = false;

      var statuses$, newStatus$;

      function showAjaxError(data) {
        var errorMessageBody = "";
        if (data.error) {
          errorMessageBody = data.error.message ? data.error.mesage : data.error;
        } else {
          $.each(data.errors, function (index, error) {
            errorMessageBody = errorMessageBody + "<li>" + error + "</li>";
          });
        }
        Message.error(MESSAGE_CONTAINER_ID, {
          title: Formatter.I18n.getText('testy.manage.status.error.title'),
          body: '<ul>' + U.htmlEscape(errorMessageBody) + '</ul>',
          fadeout: true
        });
      }

      function availableColors(id) {
        return function (colorSet) {
          var thisColor = null;
          if (id !== NEW_ID) {
            thisColor = statusMap[id].value().colorSet.toLowerCase();
          } else if (newStatusColorCell) {
            var c = newStatusColorCell.value();
            thisColor = c && c.toLowerCase();
          }
          var statuses = _.values(statusMap);
          if (creatingNew && newStatusCell.value()) {
            statuses = _.union(statuses, newStatusCell);
          }
          for (var i = 0; i < statuses.length; i++) {
            var currentSet = statuses[i].value().colorSet;
            var color = !!currentSet ? currentSet.toLowerCase() : '';
            if (color === colorSet) {
              return color === thisColor;
            }
          }
          return true;
        };
      }

      function getFreeColor(id) {
        return _.first(_.filter(AVAILABLE_SETS, availableColors(id)));
      }

      function updateColorWidth() {
        var w = 0;
        $('.tty-color-selector > .tty-ll > span').each(function () {
          w = Math.max(w, $(this).width());
        });
        colorWidth.setValue(w + 30); // 30 for padding
      }

      function propagateColorWidth(width) {
        $('.tty-color-selector > .tty-ll').width(width);
      }

      function renderColorSelector(id, name) {
        var options = _.map(_.filter(AVAILABLE_SETS, availableColors(id)), function (colorSet) {
          return {
            clazz: colorClass(colorSet),
            name: name,
            color: colorSet
          };
        });
        return manageStatusesTemplates.statusColorSelector({options: options, width: colorWidth.value()});
      }

      function closeColorSelector() {
        $(document).off('.tty-stedpanel');
        $('.tty-sted-panel').remove();
      }

      function selectColor(id, e) {
        var stedpanel$ = $(e.target).closest('.tty-sted-panel');
        var row$ = $(e.target).closest('.tty-sted-row');
        if (!stedpanel$.length || !row$.length) {
          closeColorSelector();
          return false;
        }
        closeColorSelector();
        var createNew = id === NEW_ID;
        var newColor = row$.data('color');
        var currentState, currentColor;
        if (createNew) {
          currentState = newStatusCell.value();
          currentColor = newStatusColorCell.value();
          focusInput(id);
        } else {
          currentState = statusMap[id].value();
          currentColor = currentState.colorSet;
        }
        if (newColor.toLowerCase() === currentColor.toLowerCase()) return;
        if (createNew) {
          newStatusColorCell.setValue(newColor);
        } else {
          update(_.extend({}, currentState, {colorSet: newColor}));
        }
      }

      function showColorSelector(e) {
        closeColorSelector();
        var el$ = $(e.target);
        var id = getStatusId(el$);
        var name = userStatusNameMap[id].value();
        var hideTimer;
        var offset = el$.closest('.tty-lozenge').offset();
        var selector$ = $(renderColorSelector(id, name)).css({top: offset.top - 5, left: offset.left - 12});
        $('body', document).append(selector$);
        $('.tty-sted-panel')
          .on('mouseenter', function () {
            window.clearTimeout(hideTimer);
          })
          .on('mouseleave', function () {
            hideTimer = window.setTimeout(function () { closeColorSelector(); }, 300);
            return true;
          })
          .on('click', _.partial(selectColor, id));
        return false;
      }

      function focusInput(id) {
        $('.tty-manage-statuses-container').find('[data-status-id=' + id + ']').find('input[name=name]').focus();
      }

      function addClass(attributes) {
        return $.extend({}, attributes, {
          clazz: colorClass(attributes.colorSet)
        });
      }

      function colorClass(colorSet) {
        return ('tty-' + colorSet).toLowerCase();
      }

      function setupName(row$, statusCell) {
        var status = statusCell.value();
        var view$ = row$.find('input[name=name]');
        var nameCell = new U.Cell(status.name);
        statusCell.fieldSubCell('name').pipeChanges(function (name) {
          nameCell.setValue(name);
        });
        userStatusNameMap[status.id] = nameCell;
        var readonlyText$ = row$.find('.tty-status-name-edit span');
        nameCell.pipeChanges(function (name) {
          readonlyText$.text(name);
          row$.find('.tty-status-color-edit .tty-ll > span').text(name); // Color selector is rebuilt on each status update
        });
        new U.TextPresenter(nameCell, view$);
      }

      function setupSuggestingNotes(row$, statusCell, autoUpdate) {
        var view$ = row$.find('input[name=suggestingNotes]');
        var suggestingNotesCell = statusCell.fieldSubCell('suggestingNotes');
        if (autoUpdate) {
          suggestingNotesCell.pipeChanges(function () {
            update(statusCell.value());
          });
        }
        new U.CheckboxPresenter(suggestingNotesCell, view$);
      }

      function setupNameWidth() {
        updateTextWidthLifecycle.cycle();
        _.each(_.values(userStatusNameMap), function (nameCell) {
          nameCell.pipeChanges(updateTextWidthLifecycle, updateColorWidth);
        });
        updateColorWidth();
      }

      function reloadItems() {
        TestyUtils.rest('GET', 'status', null, function (data) {
          statuses$.empty();
          newStatus$.empty();
          _.each(_.keys(statusMap), function (id) {
            statusMap[id].dispose();
            userStatusNameMap[id].dispose();
          });
          if (data && data.error) {
            showAjaxError(data);
          } else {
            if (!data || !_.isArray(data)) {
              Message.error(MESSAGE_CONTAINER_ID, {
                title: 'Error',
                body: 'Unexpected result from server: ' + U.htmlEscape(data),
                fadeout: true
              });
            } else {
              statusMap = {};
              userStatusNameMap = {};
              colorWidth.setValue(0);
              newStatusCell.setValue(null);
              var statuses = _.sortBy(data, 'priority');
              for (var i = 0; i < statuses.length; i++) {
                var status = statuses[i];
                var cell = new U.Cell(status);
                statusMap[status.id] = cell;
                var row$ = $(manageStatusesTemplates.statusRow(addClass(status)));
                setupName(row$, cell);
                setupSuggestingNotes(row$, cell, true);
                statuses$.append(row$);
              }
              var newStatusRow$ = setupNewStatusRow();
              if (newStatusRow$) {
                newStatus$.html(newStatusRow$);
              }
              setupNameWidth();
              statuses$.sortable(SORTABLE_OPTIONS);
            }
          }
        });
      }

      function setupNewStatusRow() {
        var colorSet = getFreeColor(NEW_ID);
        if (!colorSet) return null;
        var newStatus = {
          id: NEW_ID,
          name: '',
          colorSet: colorSet
        };
        if (newStatusColorCell) {
          newStatusColorCell.dispose();
        }
        newStatusColorCell = newStatusCell.fieldSubCell('colorSet').pipeChanges(updateNewStatusColor);
        newStatusCell.setValue(newStatus);
        var newStatusRow$ = $(manageStatusesTemplates.newStatus(addClass(newStatus)));
        setupName(newStatusRow$, newStatusCell);
        setupSuggestingNotes(newStatusRow$, newStatusCell, false);
        _.each(_.values(statusMap), function (status) {
          status.fieldSubCell('colorSet').pipeChanges(function () {
            newStatusColorCell.setValue(getFreeColor(NEW_ID));
          });
        });
        return newStatusRow$;
      }

      function updateNewStatusColor(colorSet) {
        $('.tty-new-status-button-container .tty-status-color-edit').html(manageStatusesTemplates.statusColor(addClass({
          name: !!userStatusNameMap[NEW_ID] ? userStatusNameMap[NEW_ID].value() : '',
          colorSet: colorSet
        })));
        propagateColorWidth(colorWidth.value());
      }

      function saveName(id) {
        var name = userStatusNameMap[id].value().trim();
        if (name.length === 0) {
          if (id === NEW_ID) {
            cancelCreatingNew();
          } else {
            cancelEdit(id, true);
          }
          return;
        }
        if (id === NEW_ID) {
          create(_.extend({}, newStatusCell.value(), {name: name}))
        } else {
          tryUpdate(id, name);
        }
      }

      function tryUpdate(id, name) {
        var state = statusMap[id].value();
        if (name === state.name) {
          cancelEdit(id, true);
        } else {
          update(_.extend({}, state, {name: name}))
        }
      }

      function update(status) {
        var row$ = getRow$(status.id);
        var loader$ = row$.find('.tty-loader-image');
        loader$.css('visibility', 'visible');
        TestyUtils.rest('PUT', 'status/' + status.id, status, function (reply) {
          loader$.css('visibility', 'hidden');
          if (reply && reply.error) {
            showAjaxError(reply);
            cancelEdit(status.id, true);
          } else {
            Message.success(MESSAGE_CONTAINER_ID, {
              title: Formatter.I18n.getText('testy.manage.status.update.success.title'),
              body: Formatter.I18n.getText('testy.manage.status.update.success.message', U.htmlEscape(status.name)),
              fadeout: true
            });
            statusMap[status.id].setValue(status);
            finishEdit(status.id);
            row$.find('.tty-status-color-edit').html(manageStatusesTemplates.statusColor(addClass(status)));
            propagateColorWidth(colorWidth.value());
          }
        });
      }

      function create(status) {
        var loader$ = getRow$(NEW_ID).find('.tty-loader-image');
        loader$.css('visibility', 'visible');
        delete status.id;
        TestyUtils.rest('POST', 'status', status, function (reply) {
          loader$.css('visibility', 'hidden');
          if (reply && reply.error) {
            showAjaxError(reply);
            cancelCreatingNew();
          } else {
            Message.success(MESSAGE_CONTAINER_ID, {
              title: Formatter.I18n.getText('testy.manage.status.create.success.title'),
              body: Formatter.I18n.getText('testy.manage.status.create.success.message', U.htmlEscape(status.name)),
              fadeout: true
            });
            creatingNew = false;
            reloadItems();
          }
        });
      }

      function remove(id) {
        if (!window.confirm(Formatter.I18n.getText('testy.manage.status.delete.confirm.text'))) return;
        var loader$ = getRow$(NEW_ID).find('.tty-loader-image');
        loader$.css('visibility', 'visible');
        TestyUtils.rest('DELETE', 'status/' + id, null, function (reply) {
          loader$.css('visibility', 'hidden');
          if (reply && reply.error) {
            showAjaxError(reply);
            cancelEdit(id, true);
          } else {
            var status = statusMap[id].value();
            Message.success(MESSAGE_CONTAINER_ID, {
              title: Formatter.I18n.getText('testy.manage.status.delete.success.title'),
              body: Formatter.I18n.getText('testy.manage.status.delete.success.message', U.htmlEscape(status.name)),
              fadeout: true
            });
            reloadItems();
          }
        });
      }

      function reorder(item) {
        var items = statuses$.find('li');
        var order = [];
        for (var i = 0; i < items.length; i++) {
          order.push($(items[i]).data('status-id'));
        }
        var loader$ = $(item).find('.tty-loader-image');
        loader$.css('visibility', 'visible');
        TestyUtils.rest('POST', '/status/reorder', order, function (reply) {
          loader$.css('visibility', 'hidden');
          if (reply && reply.error) {
            showAjaxError(reply);
            reloadItems();
          }
        });
      }

      function getRow$(id) {
        return $('#status-list').find('[data-status-id=' + id + ']');
      }

      function finishEdit(id) {
        var el$ = $('[data-status-id=' + id + '] .tty-status-name-edit.editable');
        el$.removeClass('editable').addClass('readonly');
      }

      function cancelEdit(id, clearInput) {
        finishEdit(id);
        if (!!clearInput && id !== NEW_ID) {
          userStatusNameMap[id].setValue(statusMap[id].value().name);
        }
        cancelCreatingNew();
      }

      function cancelCreatingNew() {
        if (!userStatusNameMap[NEW_ID]) return;
        userStatusNameMap[NEW_ID].setValue('');
        $('.tty-new-status-button-container').addClass('link');
        creatingNew = false;
      }

      function getStatusId(el$) {
        var id = el$.closest('.tty-status-item').data('status-id');
        return !!id ? id : NEW_ID;
      }

      function attachEventHandlers() {
        $(document).on('click', '.user-status .tty-status-name-edit.readonly', function (e) {
          if (!!getSelection().toString()) return;
          if (creatingNew) {
            cancelCreatingNew();
          }
          var nameEdit$ = $(e.target).closest('.tty-status-name-edit');
          nameEdit$.removeClass('readonly').addClass('editable');
          nameEdit$.find('input').focus();
        });
        $(document).on('keyup', function (e) {
          if (e.which === 27 || e.keyCode === 27) {
            cancelCreatingNew();
            $('.persistent .tty-status-name-edit.editable').each(function () {
              cancelEdit(getStatusId($(this)), true);
            });
          }
        });
        $(document).on('keydown', '.tty-status-name-edit', function (e) {
          if (e.which === 13 || e.keyCode === 13) {
            e.preventDefault();
            e.stopPropagation();
            saveName(getStatusId($(e.target)));
          }
        });
        $(document).on('blur', '.persistent .tty-status-name-edit.editable > input', function (e) {
          var id = getStatusId($(e.target));
          _.defer(_.partial(saveName, id));
        });

        $(document).on('click', '.user-status .tty-color-selector', showColorSelector);

        $(document).on('click', '.tty-new-status-button', function () {
          $('.tty-new-status-button-container').toggleClass('link');
          focusInput(NEW_ID);
          creatingNew = true;
          propagateColorWidth(colorWidth.value());
        });

        $(document).on('click', '.tty-status-delete > a', function (e) {
          remove(getStatusId($(e.target)));
        });
      }

      U.extend(this, {
        install: function () {
          $(function () {
            statuses$ = $('#status-list');
            newStatus$ = $('.tty-new-status');
            reloadItems();
            U.fixJQueryUISortableByAdjustingBodyOverflow();
            attachEventHandlers();
          });
        }
      });
    }

    return ManageStatuses;
  })
;
